<?php
/**
 * @file
 * Scheduled operations for classified.module
 *
 * Can be invoked from the web UI or from Drush:
 * - purge
 * - expiration
 * - notifications
 *
 * @copyright (c) 2010-2011 Ouest Systemes Informatiques (OSInet)
 *
 * @license General Public License version 2 or later
 *
 * Original code, not derived from the ed_classified module.
 */

/**
 * Unpublish nodes past their expiration date.
 *
 * The test on n.type is not required, since the inner join will only return
 * such nodes anyway, but allows the query to take advantage of the core index
 * on node.type
 *
 * Reset notify time to 0 because this field is only used for interim
 * notifications, not expire/purge.
 *
 * No db_rewrite_sql: this query can be run at any time by anyone.
 *
 * @TODO check whether the update/join syntax works on non-MySQL DB engines
 *
 * @return array
 */
function _classified_scheduled_build_expire() {
  module_load_include('inc', 'classified', 'classified_utils');

  // Obtain the list of ads to expire
  $expires = $_SERVER['REQUEST_TIME'];

  $sq = <<<EOT
    SELECT n.nid, n.title, n.uid
    FROM {node} n
      INNER JOIN {classified_node} cn ON n.nid = cn.nid
    WHERE
      n.type = 'classified' AND n.status = 1
      AND cn.expires < %d
EOT;
  $q = db_query($sq, $expires);

  $count1 = 0;
  $expired = array();
  while ($o = db_fetch_object($q)) {
    $expired[$o->uid][$o->nid] = $o->title;
    $count1++;
  }

  // Now perform the expiration.
  $sq = <<<EOT
    UPDATE {node} n
      INNER JOIN {classified_node} cn ON n.nid = cn.nid
    SET n.status = 0, cn.notify = 0
    WHERE
      n.type = 'classified' AND n.status = 1
      AND cn.expires < %d
EOT;
  $sts = db_query($sq, $expires);
  $count2 = db_affected_rows(); // should match $count1

  if ($count1 || $count2) {
    watchdog('classified', 'Expiration unpublished @count2 ads (@count1 scheduled): @expired', array(
      '@count2' => $count2,
      '@count1' => $count1,
      '@expired' => var_export($expired, TRUE),
    ),
    ($count1 == $count2) ? WATCHDOG_INFO : WATCHDOG_WARNING);
  }
  else {
    watchdog('classified', 'Expiration check did not find any ad to expire.', NULL, WATCHDOG_INFO);
  }

  drupal_alter('classified_expire', $expired);
  return $expired;
}

/**
 * Build one of the various notification lists.
 *
 * All this work can be skipped if no module implements
 * hook_classified_notify(): in such a case, notifications do not happen, and
 * there is no reason to update the notify date since notifications are not
 * being sent.
 *
 * @param string $kind
 * @return array
 */
function _classified_scheduled_build_notify($kind) {
  $notified = array();

  $modules = module_implements('classified_notify_alter');
  if (empty($modules)) {
    return $notified;
  }

  $now = $_SERVER['REQUEST_TIME'];
  $limit = $now - 12 * 60 * 60;
  switch ($kind) {
    case 'half-life':
      // cn.notify < now - 12 hours, now > n.changed + (cn.expires - n.changed)/2
      $sq = <<<EOT
        SELECT n.nid, n.uid, n.title
        FROM {node} n
          INNER JOIN {classified_node} cn ON n.nid = cn.nid
        WHERE
          n.type = 'classified' AND n.status != 0
          AND cn.notify < %d
          AND (n.changed + cn.expires) / 2 < %d
EOT;
      $sq = trim(str_replace("\n", ' ', $sq)); // clean up whitespace for regexps
      $sq = db_rewrite_sql($sq);
      $q = db_query($sq, $limit, $now);
      break;

    case 'pre-expire':
      // cn.notify > now - 12 hours, now > cn.expires - 24 hours
      $sq = <<<EOT
        SELECT n.nid, n.uid, n.title
        FROM {node} n
          INNER JOIN {classified_node} cn ON n.nid = cn.nid
        WHERE
          n.type = 'classified' AND n.status != 0
          AND cn.notify < %d
          AND cn.expires < %d
EOT;
      $sq = trim(str_replace("\n", ' ', $sq)); // clean up whitespace for regexps
      $sq = db_rewrite_sql($sq);
      $q = db_query($sq, $limit, $now + 24*60*60);
      break;

    case 'pre-purge':
      // cn.notify > now - 12 hours, now > cn.expires + grace - 24 hours
      $sq = <<<EOT
      SELECT n.nid, n.uid, n.title
        FROM {node} n
          INNER JOIN {classified_node} cn ON n.nid = cn.nid
        WHERE
          n.type = 'classified' AND n.status = 0
          AND cn.notify < %d
          AND cn.expires + %d < %d
EOT;
      $sq = trim(str_replace("\n", ' ', $sq)); // clean up whitespace for regexps
      $sq = db_rewrite_sql($sq);
      $grace = (_classified_get('grace') - 1) * 24*60*60; // 'grace' is in days
      $q = db_query($sq, $limit, $grace, $now);
      break;

    default:
      watchdog('classified', 'Invalid notify type requested: @kind',
        array('@kind' => $kind), WATCHDOG_WARNING);
      break;
  }

  $notified = array();
  while ($result = db_fetch_object($q)) {
    $notified[$result->uid][$result->nid] = $result->title;
  }

  // Alter before updating: allow modules to modify the notification list.
  drupal_alter('classified_notify', $notified, $kind);

  $updated = array();
  foreach ($notified as $uid => $user_notified) {
    foreach (array_keys($user_notified) as $nid) {
      $updated[] = $nid;
    }
  }

  // Avoid building an empty update query
  if (empty($updated)) {
    return $notified;
  }

  $placeholders = db_placeholders($updated);
  $sq = <<<EOT
    UPDATE {classified_node}
    SET notify = %d
    WHERE nid IN ($placeholders)
EOT;

  array_unshift($updated, $now); // single array format query parameter needed with a placeholders array
  $q = db_query($sq, $updated);
  $touched = db_affected_rows();
  watchdog('classified', 'Updated notification timestamp on @count ads.', array('@count' => $touched, WATCHDOG_INFO));

  return $notified;
}

/**
 * Purge nodes past their expiration date + grace period
 *
 * Selected nodes: expires + grace < now => expires < now - grace
 *
 * The test on n.type is not required, since the inner join will only return
 * such nodes anyway, but allows the query to take advantage of the core index
 * on node.type
 *
 * No db_rewrite_sql: this query can be run at any time by anyone.
 *
 * @return string
 */
function _classified_scheduled_build_purge() {
  module_load_include('inc', 'classified', 'classified_utils');

  $grace = _classified_get('grace');
  $limit = $_SERVER['REQUEST_TIME'] - $grace * 24 * 60 * 60;

  $sq = <<<EOT
    SELECT n.nid, n.title, n.uid
    FROM {node} n
      INNER JOIN {classified_node} cn ON n.nid = cn.nid
    WHERE
      n.type = 'classified' AND cn.expires < %d
EOT;
  $q = db_query($sq, $limit);

  $deleted = array();
  $count = 0;

  // Hide message information, since the page can be triggered by any user, but
  // needs to run as admin, and protect misc session content as well.
  $messages = isset($_SESSION['messages']) ? $_SESSION['messages'] : array();
  global $user;
  $saved_account = $user;
  session_save_session(FALSE);
  $user = user_load(1);
  while ($o = db_fetch_object($q)) {
    $deleted[$o->uid][$o->nid] = $o->title;
    node_delete($o->nid); // invokes drupal_set_message(), hence the hiding code
    $count++;
  }
  $user = $saved_account;
  session_save_session(TRUE);
  $_SESSION['messages'] = $messages;

  if ($count) {
    watchdog('classified', "Deleted @count nodes: @deleted", array(
      '@count' => $count,
      '@deleted' => var_export($deleted, TRUE),
    ), WATCHDOG_INFO);
  }
  else {
    watchdog('classified', 'Purge did not find any ad to delete.', NULL, WATCHDOG_INFO);
  }

  drupal_alter('classified_purge', $deleted);
  return $deleted;
}

/**
 * Page callback for expirations
 *
 * @return string
 */
function _classified_scheduled_page_expire() {
  _classified_scheduled_build_expire(); // Do not display results
  drupal_set_breadcrumb(_classified_get_breadcrumb_by_term(NULL));

  return t('<p>Thanks for triggering our expiration process.</p>');
}

/**
 * Page callback for notifications
 *
 * @param string $kind
 *
 * @return string
 */
function _classified_scheduled_page_notify($kind) {
  _classified_scheduled_build_notify($kind);
  drupal_set_breadcrumb(_classified_get_breadcrumb_by_term(NULL));

  return t('<p>Thanks for triggering our @kind notification process.</p>', array(
    '@kind' => $kind,
  ));
}

/**
 * Page callback for purges
 *
 * @return string
 */
function _classified_scheduled_page_purge() {
  _classified_scheduled_build_purge(); // Do not display results
  drupal_set_breadcrumb(_classified_get_breadcrumb_by_term(NULL));

  return t('<p>Thanks for triggering our purge process.</p>');
}
